@@ -9,7 +9,7 @@
 #include "pipewiresourcestream.h"
 #include "glhelpers.h"
 #include "logging.h"
-#include "pipewirecore.h"
+#include "pipewirecore_p.h"
 #include "pwhelpers.h"
 
 #include <libdrm/drm_fourcc.h>
@@ -30,9 +30,6 @@
 #include <EGL/egl.h>
 #include <EGL/eglext.h>
 
-#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
-#include <QtPlatformHeaders/QEGLNativeContext>
-#endif
 #undef Status
 
 #if !PW_CHECK_VERSION(0, 3, 29)
@@ -62,11 +59,13 @@ struct PipeWireSourceStreamPrivate
     spa_video_info_raw videoFormat;
     QString m_error;
     bool m_allowDmaBuf = true;
+    bool m_usingDmaBuf = false;
 
-    QHash<spa_video_format, QVector<uint64_t>> m_availableModifiers;
+    QHash<spa_video_format, QList<uint64_t>> m_availableModifiers;
     spa_source *m_renegotiateEvent = nullptr;
 
     bool m_withDamage = false;
+    Fraction maxFramerate;
 };
 
 static const QVersionNumber pwClientVersion = QVersionNumber::fromString(QString::fromUtf8(pw_get_library_version()));
@@ -90,14 +89,14 @@ uint32_t PipeWireSourceStream::spaVideoFormatToDrmFormat(spa_video_format spa_fo
     case SPA_VIDEO_FORMAT_RGB:
         return DRM_FORMAT_RGB888;
     default:
-        qCWarning(PIPEWIRE_LOGGING) << "unknown QImage format" << spa_format;
+        qCWarning(PIPEWIRE_LOGGING) << "cannot convert spa format to fourcc" << spa_format;
         return DRM_FORMAT_INVALID;
     }
 }
 
-static QHash<spa_video_format, QVector<uint64_t>> queryDmaBufModifiers(EGLDisplay display, const QVector<spa_video_format> &formats)
+static QHash<spa_video_format, QList<uint64_t>> queryDmaBufModifiers(EGLDisplay display, const QList<spa_video_format> &formats)
 {
-    QHash<spa_video_format, QVector<uint64_t>> ret;
+    QHash<spa_video_format, QList<uint64_t>> ret;
     ret.reserve(formats.size());
     const bool hasEglImageDmaBufImportExt = epoxy_has_egl_extension(display, "EGL_EXT_image_dma_buf_import");
     static auto eglQueryDmaBufModifiersEXT = (PFNEGLQUERYDMABUFMODIFIERSEXTPROC)eglGetProcAddress("eglQueryDmaBufModifiersEXT");
@@ -106,12 +105,12 @@ static QHash<spa_video_format, QVector<uint64_t>> queryDmaBufModifiers(EGLDispla
     EGLint count = 0;
     EGLBoolean successFormats = eglQueryDmaBufFormatsEXT(display, 0, nullptr, &count);
 
-    QVector<uint32_t> drmFormats(count);
+    QList<uint32_t> drmFormats(count);
     successFormats &= eglQueryDmaBufFormatsEXT(display, count, reinterpret_cast<EGLint *>(drmFormats.data()), &count);
     if (!successFormats)
         qCWarning(PIPEWIRE_LOGGING) << "Failed to query DMA-BUF formats.";
 
-    const QVector<uint64_t> mods = hasEglImageDmaBufImportExt ? QVector<uint64_t>{DRM_FORMAT_MOD_INVALID} : QVector<uint64_t>{};
+    const QList<uint64_t> mods = hasEglImageDmaBufImportExt ? QList<uint64_t>{DRM_FORMAT_MOD_INVALID} : QList<uint64_t>{};
     if (!eglQueryDmaBufFormatsEXT || !eglQueryDmaBufModifiersEXT || !hasEglImageDmaBufImportExt || !successFormats) {
         for (spa_video_format format : formats) {
             ret[format] = mods;
@@ -139,7 +138,7 @@ static QHash<spa_video_format, QVector<uint64_t>> queryDmaBufModifiers(EGLDispla
             break;
         }
 
-        QVector<uint64_t> modifiers(count);
+        QList<uint64_t> modifiers(count);
         if (count > 0) {
             if (!eglQueryDmaBufModifiersEXT(display, drm_format, count, modifiers.data(), nullptr, &count)) {
                 qCWarning(PIPEWIRE_LOGGING) << "Failed to query DMA-BUF modifiers.";
@@ -203,7 +202,8 @@ void PipeWireSourceStream::renegotiateModifierFailed(spa_video_format format, qu
     pw_loop_signal_event(d->pwCore->loop(), d->m_renegotiateEvent);
 }
 
-static spa_pod *buildFormat(spa_pod_builder *builder, spa_video_format format, const QVector<uint64_t> &modifiers, bool withDontFixate)
+static spa_pod *
+buildFormat(spa_pod_builder *builder, spa_video_format format, const QList<uint64_t> &modifiers, bool withDontFixate, const Fraction &requestedMaxFramerate)
 {
     spa_pod_frame f[2];
     const spa_rectangle pw_min_screen_bounds{1, 1};
@@ -214,12 +214,19 @@ static spa_pod *buildFormat(spa_pod_builder *builder, spa_video_format format, c
     spa_pod_builder_add(builder, SPA_FORMAT_mediaSubtype, SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw), 0);
     spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_format, SPA_POD_Id(format), 0);
     spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_size, SPA_POD_CHOICE_RANGE_Rectangle(&pw_min_screen_bounds, &pw_min_screen_bounds, &pw_max_screen_bounds), 0);
+    if (requestedMaxFramerate) {
+        auto defFramerate = SPA_FRACTION(0, 1);
+        auto minFramerate = SPA_FRACTION(1, 1);
+        auto maxFramerate = SPA_FRACTION(requestedMaxFramerate.numerator, requestedMaxFramerate.denominator);
+        spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_framerate, SPA_POD_Fraction(&defFramerate), 0);
+        spa_pod_builder_add(builder, SPA_FORMAT_VIDEO_maxFramerate, SPA_POD_CHOICE_RANGE_Fraction(&maxFramerate, &minFramerate, &maxFramerate), 0);
+    }
 
     if (modifiers.size() == 1 && modifiers[0] == DRM_FORMAT_MOD_INVALID) {
         // we only support implicit modifiers, use shortpath to skip fixation phase
         spa_pod_builder_prop(builder, SPA_FORMAT_VIDEO_modifier, SPA_POD_PROP_FLAG_MANDATORY);
         spa_pod_builder_long(builder, modifiers[0]);
-    } else if (modifiers.size()) {
+    } else if (!modifiers.isEmpty()) {
         // SPA_POD_PROP_FLAG_DONT_FIXATE can be used with PipeWire >= 0.3.33
         if (withDontFixate) {
             spa_pod_builder_prop(builder, SPA_FORMAT_VIDEO_modifier, SPA_POD_PROP_FLAG_MANDATORY | SPA_POD_PROP_FLAG_DONT_FIXATE);
@@ -258,9 +265,10 @@ void PipeWireSourceStream::onStreamParamChanged(void *data, uint32_t id, const s
     // the server announces support for it.
     // See https://github.com/PipeWire/pipewire/blob/master/doc/dma-buf.dox
 
-    const auto bufferTypes = pw->d->m_allowDmaBuf && spa_pod_find_prop(format, nullptr, SPA_FORMAT_VIDEO_modifier)
-        ? (1 << SPA_DATA_DmaBuf) | (1 << SPA_DATA_MemFd) | (1 << SPA_DATA_MemPtr)
-        : (1 << SPA_DATA_MemFd) | (1 << SPA_DATA_MemPtr);
+    pw->d->m_usingDmaBuf = pw->d->m_allowDmaBuf && spa_pod_find_prop(format, nullptr, SPA_FORMAT_VIDEO_modifier);
+    Q_ASSERT(pw->d->m_allowDmaBuf || !pw->d->m_usingDmaBuf);
+    const auto bufferTypes =
+        pw->d->m_usingDmaBuf ? (1 << SPA_DATA_DmaBuf) | (1 << SPA_DATA_MemFd) | (1 << SPA_DATA_MemPtr) : (1 << SPA_DATA_MemFd) | (1 << SPA_DATA_MemPtr);
 
     QVarLengthArray<const spa_pod *> params = {
         (spa_pod *)spa_pod_builder_add_object(&pod_builder,
@@ -334,7 +342,8 @@ PipeWireSourceStream::PipeWireSourceStream(QObject *parent)
     : QObject(parent)
     , d(new PipeWireSourceStreamPrivate)
 {
-    qRegisterMetaType<QVector<DmaBufPlane>>();
+    qRegisterMetaType<QList<DmaBufPlane>>();
+    qRegisterMetaType<PipeWireCursor>();
 
     pwStreamEvents.version = PW_VERSION_STREAM_EVENTS;
     pwStreamEvents.process = &onProcess;
@@ -362,22 +371,41 @@ Fraction PipeWireSourceStream::framerate() const
     return {0, 1};
 }
 
+void PipeWireSourceStream::setMaxFramerate(const Fraction &framerate)
+{
+    d->maxFramerate = framerate;
+
+    if (d->pwStream) {
+        pw_loop_signal_event(d->pwCore->loop(), d->m_renegotiateEvent);
+    }
+}
+
 uint PipeWireSourceStream::nodeId()
 {
     return d->pwNodeId;
 }
 
-QVector<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_builder podBuilder)
+QList<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_builder podBuilder)
 {
     const auto pwServerVersion = d->pwCore->serverVersion();
-    const QVector<spa_video_format> formats =
-        {SPA_VIDEO_FORMAT_RGBx, SPA_VIDEO_FORMAT_RGBA, SPA_VIDEO_FORMAT_BGRx, SPA_VIDEO_FORMAT_BGRA, SPA_VIDEO_FORMAT_RGB, SPA_VIDEO_FORMAT_BGR};
-    QVector<const spa_pod *> params;
+    const QList<spa_video_format> formats = {
+        SPA_VIDEO_FORMAT_RGBx,
+        SPA_VIDEO_FORMAT_RGBA,
+        SPA_VIDEO_FORMAT_BGRx,
+        SPA_VIDEO_FORMAT_BGRA,
+        SPA_VIDEO_FORMAT_RGB,
+        SPA_VIDEO_FORMAT_BGR,
+        SPA_VIDEO_FORMAT_xBGR,
+        SPA_VIDEO_FORMAT_BGRA,
+        SPA_VIDEO_FORMAT_ABGR,
+        SPA_VIDEO_FORMAT_xBGR,
+    };
+    QList<const spa_pod *> params;
     params.reserve(formats.size() * 2);
     const EGLDisplay display = static_cast<EGLDisplay>(QGuiApplication::platformNativeInterface()->nativeResourceForIntegration("egldisplay"));
 
-    d->m_allowDmaBuf = pwServerVersion.isNull() || (pwClientVersion >= kDmaBufMinVersion && pwServerVersion >= kDmaBufMinVersion);
-    const bool withDontFixate = pwServerVersion.isNull() || (pwClientVersion >= kDmaBufModifierMinVersion && pwServerVersion >= kDmaBufModifierMinVersion);
+    d->m_allowDmaBuf = d->m_allowDmaBuf && (pwServerVersion.isNull() || (pwClientVersion >= kDmaBufMinVersion && pwServerVersion >= kDmaBufMinVersion));
+    const bool withDontFixate = d->m_allowDmaBuf && (pwServerVersion.isNull() || (pwClientVersion >= kDmaBufModifierMinVersion && pwServerVersion >= kDmaBufModifierMinVersion));
 
     if (d->m_availableModifiers.isEmpty()) {
         d->m_availableModifiers = queryDmaBufModifiers(display, formats);
@@ -385,10 +413,10 @@ QVector<const spa_pod *> PipeWireSourceStream::createFormatsParams(spa_pod_build
 
     for (auto it = d->m_availableModifiers.constBegin(), itEnd = d->m_availableModifiers.constEnd(); it != itEnd; ++it) {
         if (d->m_allowDmaBuf && !it->isEmpty()) {
-            params += buildFormat(&podBuilder, it.key(), it.value(), withDontFixate);
+            params += buildFormat(&podBuilder, it.key(), it.value(), withDontFixate, d->maxFramerate);
         }
 
-        params += buildFormat(&podBuilder, it.key(), {}, withDontFixate);
+        params += buildFormat(&podBuilder, it.key(), {}, withDontFixate, d->maxFramerate);
     }
     return params;
 }
@@ -446,6 +474,7 @@ void PipeWireSourceStream::handleFrame(struct pw_buffer *buffer)
         using namespace std::chrono;
         auto now = system_clock::now();
         d->m_currentPresentationTimestamp = time_point_cast<nanoseconds>(now).time_since_epoch();
+        frame.presentationTimestamp = d->m_currentPresentationTimestamp;
     }
 
     if (spa_meta *vd = spa_buffer_find_meta(spaBuffer, SPA_META_VideoDamage)) {
@@ -472,8 +501,6 @@ void PipeWireSourceStream::handleFrame(struct pw_buffer *buffer)
                     PWHelpers::SpaBufferToQImage(bitmap_data, bitmap->size.width, bitmap->size.height, bitmap->stride, spa_video_format(bitmap->format));
             }
             frame.cursor = {{cursor->position.x, cursor->position.y}, {cursor->hotspot.x, cursor->hotspot.y}, cursorTexture};
-        } else {
-            frame.cursor = {{}, {}, {}};
         }
     }
 
@@ -563,3 +590,20 @@ void PipeWireSourceStream::setDamageEnabled(bool withDamage)
 {
     d->m_withDamage = withDamage;
 }
+
+bool PipeWireSourceStream::usingDmaBuf() const
+{
+    return d->m_usingDmaBuf;
+}
+
+bool PipeWireSourceStream::allowDmaBuf() const
+{
+    return d->m_allowDmaBuf;
+}
+
+void PipeWireSourceStream::setAllowDmaBuf(bool allowed)
+{
+    d->m_allowDmaBuf = allowed;
+}
+
+#include "moc_pipewiresourcestream.cpp"
